# dashc, client emulator for DASH video streaming

## Description

dashc is a video player emulator, it is intended to be used as a test tool. The original goal was to have a lightweight test tool for network performance experiments with a real DASH video streaming traffic.

dashc has several options which can be changed through the command line arguments:

* video adaptation algorithm (Conventional, BBA-0/BBA-1/BBA-2, ARBITER);
* initial buffer size (in segments);
* last segment index for play;
* log file and folder name (the log file can be on/off). The log file will be stored with other information in a directory named with the current timestamp;
* maximum buffer size (in seconds);
* persistent connection (on/off);
* generation of the meta information for BBA-2 and ARBITER (file with segment sizes);
* measurement of ITU P.1203 Quality of Experience

## Installation

dashc had been used in Ubuntu 16.04/16.10/17.04/17.10/18.04 with x64 architecture (continuously for an ~1 year), in Ubuntu 16.04.3 with Raspberry Pi 2/3 (briefly), in macOS High Sierra 10.13 (briefly).

To install (./configure might ask for a sudo password to install opam, m4, git):

```bash
./configure
make
```

To run tests (not necessary):

```bash
make test
```

## Usage

dashc can be run as simple as:

```bash
./dashc.exe play http://10.0.0.1/bbb.mpd
```

or:

```bash
./dashc.exe play http://10.0.0.1/bbb.mpd [-adapt conv] [-initb 2] [-maxbuf 60] [-persist true] [-turnlogon true] [-logname now] [-subfolder qwe] [-extended false] [-no-segs 9999999999] [-preferred-codec none]
```

Where all flags except the link to the MPD file are optional. Possible flags are:

* [-adapt], adaptation algorithm (conv; bba-0; bba-1; bba-2; arbiter);
* [-initb], initial buffer (in segments);
* [-lastsegmindex], last segment index for play;
* [-logname], name of the log file ("now" means usage of the current time stamp and it is used by default);
* [-maxbuf], maximum buffer size (in seconds);
* [-persist], persistent connection (true/false);
* [-segmentlist], this parameter is used to tell where to get meta data infromation (segment sizes, for example, for BBA-2 and ARBITER). Possible options are head (dashc will send head requests before starting to stream), local (file with segment size should be at the same folder with dashc), remote (file with segment size should be at the same location where target MPD file is stored) get segment sizes from  local - local segmentlist_%mpd_name%.txt file. The default option is head;
* [-subfolder], subfolder for the file;
* [-turnlogon], turn on logging to file;
* [-gensegmfile], generate segmentlist_%mpd_name%.txt file only (it will be rewritten if exists) - input values [false, true] - false by default;
* [-extended], display aditional columns with information for producing QOE metrics;
* [-help],  print this help text and exit, (alias: -?);
* [-preferred-codec], (h264, h265, vp09, av01) stream using the specified codec. Will raise an error and terminate execution if the codec is unavailable in the specified mpd file. Defaults as ```bash -preferred-codec none```
* [-no-segs] the maximum number of segments to stream (defaults as 9999999999)
* [-max-rep-height] the maximum resolution height to stream at (defauls as 2160)

Example to generate the segment list file
```bash
./dashc.exe play http://10.0.0.1/bbb.mpd -gensegmfile true
```

Example to use the segment list file
./dashc.exe play http://10.0.0.1/bbb.mpd -segmentlist local -adapt bba-2


The detailed help can be found by running it with the -help flag:

```bash
./dashc.exe play -help
```

The typical log file will look like the example below (here the data only for the first 30 segments is shown, ARBITER adaptation algorithm is used, bbb-264 4 seconds segment size clip is used):

```bash
Seg_#  Arr_time  Del_Time  Stall_Dur  Rep_Level  Del_Rate  Act_Rate  Byte_Size  Buff_Level  RTT
    1        66        66    0.000000        232     14862       248    124131     4.000  0.003071
    2       123        53    0.000000        232     20821       276    138452     8.000  0.038723
    3       566       404    0.000000       4275     52658      5323   2661696    11.596  0.003084
    4       754       184    0.000000       4275     69018      3187   1593595    15.411  0.003223
    5       864       107    0.000000       4275     56887      1524    762041    19.304  0.004062
    6      1076       208    0.000000       4275     75145      3911   1955719    23.096  0.003290
    7      1477       397    0.000000       4275     78086      7761   3880949    26.698  0.003215
    8      1776       295    0.000000       4275     75233      5555   2777836    30.403  0.003927
    9      1970       189    0.000000       4275     74129      3517   1758687    34.213  0.005750
   10      2129       153    0.000000       4275     53808      2061   1030814    38.059  0.003664
   11      2265       131    0.000000       4275     69041      2276   1138183    41.928  0.003619
   12      2460       191    0.000000       4275     76517      3670   1835157    45.736  0.003283
   13      2718       254    0.000000       4275     80125      5099   2549858    49.481  0.003315
   14      2954       232    0.000000       4275     79446      4623   2311934    53.248  0.003257
   15      3054        96    0.000000       4275     61330      1476    738026    57.152  0.004894
   16      4457       196    0.000000       4275     73192      3603   1801745    59.755  0.004198
   17      8510       243    0.000000       4275     71802      4362   2181048    59.712  0.003321
   18     12516       197    0.000000       4275     73164      3608   1804320    59.715  0.003850
   19     16912       586    0.000000       4275     38334      5622   2811097    59.328  0.003742
   20     20760       440    0.000000       4275     82178      9040   4520157    59.488  0.003835
   21     24416       147    0.000000       4275     69034      2539   1269826    59.841  0.003753
   22     28425        99    0.000000       4275     58030      1448    724180    59.841  0.003859
   23     32572       238    0.000000       4275     51038      3038   1519370    59.703  0.003893
   24     36495       113    0.000000       4275     62611      1777    888939    59.789  0.003937
   25     40481       176    0.000000       4275     72914      3210   1605463    59.813  0.003825
   26     44618       228    0.000000       4275     77702      4434   2217356    59.685  0.003747
   27     48601       274    0.000000       4275     65617      4496   2248266    59.711  0.003815
   28     52635       225    0.000000       4275     75694      4261   2130911    59.686  0.003897
   29     56653       308    0.000000       4275     68749      5306   2653059    59.677  0.010428
   30     61026       656    0.000000       4275     54226      8906   4453235    59.320  0.004737
```

The columns include the following information:

* segment number starting from 1;
* arrival time of a segment in milliseconds;
* time spent for delivery of this segment in milliseconds;
* stall (freeze) duration in seconds;
* representation rate of downloaded segment in Kbit/s (taken from MPD file);
* delivery rate of the network in Kbit/s (segment size divided by time for delivery);
* the actual bitrate of this segment (segment size divided by the segment duration) in Kbit/s;
* segment size in bytes;
* buffer level after this segment was just downloaded, in seconds;
* the approximate RTT (round trip time) in seconds. Note this this may be marginally higher than expected as it is based on download times for the mpd file;

The file will appear as follows when the -extended parameter is true

```bash
WARNING: WidthxHeigth (3840x2160) of clip @ rate 39.54Mbps - is greater than defined max_rep height 1080
WARNING: WidthxHeigth (3840x2160) of clip @ rate 24.72Mbps - is greater than defined max_rep height 1080
WARNING: WidthxHeigth (3840x2160) of clip @ rate 14.82Mbps - is greater than defined max_rep height 1080
Seg_#  Arr_time  Del_Time  Stall_Dur  Rep_Level  Del_Rate  Act_Rate  Byte_Size  Buff_Level  RTT       Codec   Width   Height   FPS   Seg_Dur   Start   P.1203 QoE
    1        52        52    0.000000        237      7421        98     49093     4.000  0.038965     h264     320     180    60     4.000     0.000   1.936882
    2       793       100    0.000000        237     16830       424    212181     8.000  0.038194     h264     320     180    60     4.000     4.000   1.936882
    3      4491       953    0.000000       4282     25941      6180   3090357     9.900  0.036395     h264    1920    1080    60     4.000     8.000   2.916288
    4      9774       638    0.000000       4282     25003      3993   1996631     9.900  0.038734     h264    1920    1080    60     4.000    12.000   3.257684
    5     15087       659    0.000000       4282     24027      3964   1982069     9.900  0.037138     h264    1920    1080    60     4.000    16.000   3.433467 
```

The extra columns include the information:

* codec used to encode each segment (h264, h265, vp9, av1);
* width and;
* height of segment being streamed in pixels;
* framerate in frames per second;
* the duration (in seconds) of each segment;
* the start time of each segment, in seconds, relative to the start time of the first segment;
* the PREVIOUS segment's QoE (on a scale of 1-5) which is calculated using the ITU P.1203 tool in mode 0 (https://github.com/itu-p1203/itu-p1203).
NOTE: the QoE for the final segment is stored in the log directory (e.g. ./2018-08-02_12-03-34849149/) for that run in a file named temp.tmp

After the stream has completed, dashc will output a summary of the above metadata along with any warnings encountered to the standard output and to a separate logfile (located in the log directory) named summary.log

```bash
Downloaded 5 segments (4-second h264) totalling    7.16 MB
P.1203 Quality of Experience:			  3.4335
Round Trip Time (RTT) min/avg/max:  0.036/0.038/0.039
Delivery rate min/avg/max:		      7421.186/19844.400/25941.238 KB/s
Number of stalls/Total stall time:  0/0.000
Representation level min/avg/max:		237.851/2664.541/4282.335 KB/s
Resolution min/mode/max:			      320x180/1920x1080/1920x1080 
```

This summary will display:

* the number, codec, duraton and overall size in megabytes of segments downloaded;
* the final QoE value for the stream (P.1203 mode 0), if applicable;
* the minimum, average and maximum round trip times for the entire stream in seconds;
* the minimum, average and maximum  delivery rate in kilobytes per second;
* the number of stalls and their combined duration in seconds;
* the minimum, average and maximum representation levels in kilobits per second;
* the minimum, modal and maximum resolutions used;

## Log Files and Directories

For every run, Dashc creates a log directory in the dashc directory, named with the current timestamp. In here the log file, downloaded segment files and an ITU P.1203 mode 0 compatible json file (temp.json) are stored along with other temporary files. These directories can potentially contain large amounts of data (multiple GB) depending on the streams used, so it is important to regularly delete these directories and their contents in order to avoid storage issues on your device. This can be accomplished as easily as:

```bash
rm -R ~/dashc/2*
```

## Warnings

In some cases, dashc will display warnings to the standard output (but not the log file):

if a clip resolution is larger than a defined pixel height (this is used to keep a UHD clip within the FUll HD range supported by P.1203)

```bash
WARNING: WidthxHeigth (3840x2160) of clip @ rate 39.54Mbps - is greater than defined max_rep height 1080
```

If an mpd file is specified with resolutions above full HD (1920x1080) the following warning will be displayed.

```bash
WARNING: this stream contains resolutions greater than full HD (1920x1080) which are not supported by the ITU P.1203 standard
```
If Dashc downloads a segment with a resolution greater than full HD the following warning will be shown, and QoE values will no longer be calculated.

```bash
WARNING: codec not compatible with P.1203. QoE values may not be accurate.
```

If a codec other than h264 is used, Dashc will issue the follwing warning and stop calculating QoE values.

```bash
WARNING: codec not compatible with P.1203
```

All warnings will be displayed to the standard output only, not to the log file.
All warnings will be re-rdisplayed after the stream has completed with the summary statistics and will be stored in the summary file.

## Dataset

The tool was tested with this dataset https://www.ucc.ie/en/misl/research/datasets/ivid_dataset/. The example of the log file above was taken with the Big Buck Bunny movie 4 second segment size x264 video codec and caddy web server (<https://en.wikipedia.org/wiki/Caddy_(web_server)>).

## Multi-codec MPD file support

Dashc supports multi-codec MPD files such as those at (<http://www-itec.uni-klu.ac.at/ftp/datasets/mmsys18/>) compiled by Anatoliy Zabrovskiy, Christian Feldmann and Christian Timmerer, although currently only h264 is supported by dashc for this dataset.

If dashc encounters a multi-codec MPD and a codec has been specified with the preferred-codec argument, it will stream using this codec. If no codec has been specified, it will use the last defined adaption set in the mpd file.

Currently, dashc can recognise h264, h265, vp9 and av1 codecs.


## Support

If you have an issue, please create a new one in github with necessary information (type of MPD file, OS version and etc.).

## FAQ

Q: X package (dependency) cannot be installed.

A: Update of the packages database might help.

```bash
opam update          # Update the packages database
opam upgrade         # Bring everything to the latest version possible
```

Q: After the first ./configure, the make prints something like:

```bash
make: jbuilder: Command not found
Makefile:4: recipe for target 'all' failed
make: *** [all] Error 127
```

A: Log out of the OS and log in back might help.

Q: Why there was a need in a new implementation instead of updating already existing similar solutions?

A: It was clear that the new implementation won't take much time.

Q: Why OCaml was used (not a C/C++/Java/C#/Python)?

A: OCaml allowed to make a working tool in several days due to hands-on expirience with it.

Q: What is OCaml?

A: A good set of resources can be found here:
* <https://ocaml.org/learn/>
* <https://github.com/rizo/awesome-ocaml>
* <https://github.com/vramana/awesome-reasonml>
* <https://dev.realworldocaml.org/>
* <http://www.cs.cornell.edu/courses/cs3110>
* <http://alan.petitepomme.net/cwn/> (OCaml weekly news)
* <https://discuss.ocaml.org/>
* Some videos about OCaml (<https://www.youtube.com/watch?v=v1CmGbOGb2I>, <https://www.youtube.com/watch?v=-J8YyfrSwTk>, <https://www.youtube.com/watch?v=FnBPECrSC7o>)

Q: Are there any good IDEs for OCaml?n

A: Visual Studio Code with reasonml plugin  (<https://marketplace.visualstudio.com/items?itemName=freebroccolo.reasonml>) is good. Merlin and ocp-indent should be installed as well:
```bash
opam install merlin ocp-indent
```

Q: Are there any open source code examples in OCaml?

A: Some open source projects:
* <http://www.cs.cornell.edu/courses/cs3110/archive/tournaments.html>
* <https://github.com/facebook/infer>
* <https://github.com/facebook/flow>
* <https://github.com/facebook/pyre-check>


## Build errors
1.  Some errors with ocaml, can be fixed by removing ocaml or unpinning some ocaml modules:
sudo apt-get purge ocaml
opam pin remove -y cohttp ocaml-cohttp
opam remove -y cohttp cohttp-async
opam pin remove -y cohttp-async ocaml-cohttp


2.  A build error noticed for P.1203:
when calling "python3 -m itu_p1203 ./itu-p1203/examples/mode0.json"
AttributeError: type object 'numpy.ndarray' has no attribute '__array_function__'
to fix:
pip3 uninstall numpy
pip3 install numpy==1.15.4
pip3 uninstall numpy
pip3 install numpy

## Citation

Aleksandr Reviakin, Ahmed H. Zahran, Cormac J. Sreenan. dashc : a highly scalable client emulator for DASH video. MMSys 2018 Open Dataset & Software Track.
Andrew Nash, Benjamin Provan-Bessell. MISL Summer of Code 2018

## Licence

This software has emanated from research conducted with the financial support of Science Foundation Ireland (SFI) under Grant Number 13/IA/1892.

University College Cork

GPL-2 License
