#! /bin/bash
#
#	dashc, client emulator for DASH video streaming
#	Copyright (c) 2016-2018, Aleksandr Reviakin, University College Cork
#                      2018, Andrew Nash, Benjamin Provan-Bessell, MISL Summer of Code 2018
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#	02110-1301, USA.
#

file_structure=$1
codec=$2
duration=$3
fps=$4
resolution=$5
start=$6
in_number=$7
stalling_time=$8
basecase="0."
dr=${file_structure%/*}
filesize=$(stat -c%s $file_structure) #size in bytes

#there will be 3 files created before a final json file can be formed
#head.json will store the header and the segment data
#tail.json will store the stall data
#tailend.json will store the footer 
file="$dr/head.json" 
mdatValue=`LANG=C grep -obUaP "\x00\x00\x00\x04\x68\xEF\xBC\x80" $file_structure | awk 'BEGIN{FS=":"}{print $1}'`
mdatValue=`echo $mdatValue1 | awk '{print $1}'`
header_val=$(($mdatValue + 8))
without_header_val=$(($filesize-$header_val))
bitrate=$(($without_header_val*2))
#printf -v will store the output in variable kbps
printf -v kbps '%.3f\n' "$(echo "scale=2; $bitrate/1024" | bc)"

#init is the time the stream started stalling
#start is the time the stream resumed after stalling
init=`echo "$start - $stalling_time" | bc`
#check if the head.json file exists 
#this is to ensure that segment data are correctly separated by printing a comma before a segment if it is not the first segment to be written to the file
if [ -e "$file" ]
then 
    #append the relavent data to the head.json file
    echo -e ",{\"bitrate\": $kbps,\n\"codec\": \"$codec\",\n\"duration\": ${duration}0,\n\"fps\": $fps,\n\"resolution\": \"$resolution\",\n\"start\": ${start}0\n}" >> "$file"
    #check if this segment stalled
    if [ "$stalling_time" != "$basecase" ];
        then 
        #we need to comma separate stall data without and leading or trailing commas
        #if the last character in the tail.json file is [ then this is the first element to be added and we should not print a comma
        #otherwise, stall data has already been added and we should print a comma
        last=`tail -c 2 "$dr/tail.json"`
        pos="["
        if [ "$last" != "$pos" ];
            then echo "," >> "$dr/tail.json"
        fi
        #append the stall data to tail.json
        #0 is added after init as the floating value was sometimes formatted as x.  (e.g. 5. ) which caused issues with the P.1203 module
        echo "["${init}0","${stalling_time}"]" >> "$dr/tail.json"
    fi
else
    #create the appropiate headers in the head and tail files and the appropiate footer in the tailend file
    echo -e "{\"I11\":{\"segments\": []}, \"I13\": {\"segments\": [\n{\n\"bitrate\": $kbps,\n\"codec\": \"$codec\",\n\"duration\": ${duration}0,\n\"fps\": $fps,\n\"resolution\": \"$resolution\",\n\"start\": ${start}0\n}" > "$file"
    echo -e "]},\"I23\": {\"stalling\": [" > "$dr/tail.json"
    echo -e "]},\"IGen\": {\"device\": \"pc\",\"displaySize\": \"1920x1080\",\"viewingDistance\": \"150cm\"}}" >  "$dr/tailend.json"
fi
#combine the parts into a final mode 0 compatible json file
cat "$file" "$dr/tail.json" "$dr/tailend.json" > "$dr/temp.json"
#run itu p.1203 (sending warnings/errors to /dev/null, effectively ignoring them) and extract the qoe only from the output
#python3 -m itu_p1203 $dr/temp.json
qoe=$(python3 -m itu_p1203 $dr/temp.json 2> /dev/null | tail -n 6 | head -n1 | cut -f 1 -d ',' | cut -f 4 -d ' ')
echo "$qoe"
